#define PTree boost::shared_ptr<class CTree>
#define PRef  boost::shared_ptr<class CRef>

//! Context for generic tree evaluation.

class CContext {
 public:
    enum TReqDim { // Dimensionality of requested data
        _1,  // scalar value
        _VI, // vector valued as function of data index _I
        _VT  // vector valued as function of function argument t
    };
    enum TReqTyp { // Type of requested data
        _QI,  // index
        _QD   // double
    };
    TReqDim reqDim;
    uint k, j, i;
    const vector<double> *fargs;
    uint nv;
    inline CContext( const uint k_=-1, const uint j_=-1, const uint i_=-1 ):
        k(k_), j(j_), i(i_), reqDim(_1), fargs(0) {};
    void set_fargs( const vector<double> *fargs_ ) {
        fargs = fargs_; reqDim = _VT; }
    void set_nv( const uint nv_ ) { nv = nv_; reqDim = _VI; }

    bool dim_1()  const { return reqDim==_1;  };
    bool dim_vi() const { return reqDim==_VI; };

    string info() const;

};


//! Curve reference with evaluation context.

typedef struct {
    const class COlc* fc;
    uint j, k;
} CurveRef;


//! Return value for tree evaluation.

class CTOut {
    enum TOutTyp { _OI, _OD, _OF, _OVD };
 public:
    TOutTyp typ;
    // int u;
    double d;
    vector<double> vd;
    CurveRef cr;
    inline void set_u( const int u_ ) { typ = _OI; d = u_; };
    inline void set_d( const double d_ ) { typ = _OD; d = d_; };
    inline void preset_vd( const int n ) { typ = _OVD; vd.resize(n); };
    inline void set_vd( const vector<double>& vd_ ) { typ = _OVD; vd = vd_; };
    inline void set_cr( const class COlc* fc_, const uint k_, const uint j_ ) {
        typ = _OF; cr.fc = fc_; cr.k = k_; cr.j = j_; };
    inline bool is_scalar() const { return typ==_OI || typ==_OD; };
    inline bool is_vector() const { return typ==_OVD; };
    inline bool is_curveref() const { return typ==_OF; };
    inline double to_d( const int i ) const { return is_scalar() ? d : vd[i]; }
    inline void to_vd( vector<double> *vd_out ) const {
        if( is_scalar() ) {
            vd_out->resize( 1 );
            (*vd_out)[0];
        } else
            (*vd_out = vd);
    }
    string info() const;
};


//! References to data points, indices, ...; used as endpoints of CTree's.

class CRef {
 public:
    enum TVar { _NOREF=0,
                _X, _Y, _Z,    // data
                _FC, _FP, _FM, // "fit curve", function par's, fit metrics
                _K, _J, _I,    // indices for file, scan, point   
                _NJ, _NI       // number of scans in file, points in scan
    };
    TVar var;
    uint num;
    PTree tk;
    PTree tj;
    PTree ti;

    CRef( const string s="",
          PTree _tk=PTree(), PTree _tj=PTree(), PTree _ti=PTree());

    inline bool defined() const {
        return var!=_NOREF; };
    inline bool isindex() const {
        return var==_K || var==_J || var==_I; };
    inline bool issize() const {
        return var==_NJ || var==_NI; };
    inline bool pointwise() const { 
        return var==_X || var==_Y || var==_I; };
    inline bool scanwise() const {
        return pointwise() || var==_Z || var==_FP || var==_FM || var==_J ||
            var==_NI || var==_FC; };
    inline bool unspecific() const { 
        return isindex() || issize(); };
    inline bool indatafile() const {
        return unspecific() || var==_X || var==_Y || var==_Z; };
    inline bool incurvefile() const {
        return (unspecific() && !pointwise())
            || var==_FC || var==_FP || var==_FM || var==_Z ; };
    inline bool numbered() const {
        return var==_FP || var==_FM || var==_Z; };
    inline bool assignable() const {
        return defined() && !unspecific(); };
    inline bool indexed() const {
        return ti || tj || tk; };
    
    void ref_val( CTOut* ret, const CContext *ctx ) const;
    void coord( CCoord *ret, uint k ) const;
    string ref_info() const;
    inline const char* ref_cstr() const { return ref_info().c_str(); };
};


//! Parsed generic expression.

class CTree {
 private:
    enum TreeTyp { _NOTREE=0,
                   _OP,   // operator (unary, binary, .. according to narg)
                   _VAL,  // numeric value
                   _FRF,  // function reference 'f'
                   _FARG, // function argument 't'
                   _REF,  // reference
                   _CONV, // convolution
                   _DIRAC // Dirac delta function
    };
    TreeTyp typ;
    double val;
    const CFunc* fun;
    static const uint maxarg=3;
    PTree arg[maxarg];
    uint narg;
    PRef ref;

 public:
    bool has_farg;

    inline CTree()
        : typ(_NOTREE), fun(0), has_farg(false) {};
    inline CTree( double v_in )
        : typ(_VAL), val(v_in), fun(0), has_farg(false) {};
    inline CTree( const CFunc *f_in, PTree a0 )
        : typ(_OP), narg(1), fun(f_in) {
        arg[0] = a0;
        has_farg = a0->has_farg;
        if ( fun->narg!=1 )
            throw string( "PROG ERR: CTree(1) with conflicting #arg" );
    };
    inline CTree( const CFunc *f_in, PTree a0, PTree a1 )
        : typ(_OP), narg(2), fun(f_in) {
        arg[0] = a0; arg[1] = a1;
        has_farg = a0->has_farg || a1->has_farg;
        if ( fun->narg!=2 )
            throw string( "PROG ERR: CTree(2) with conflicting #arg" );
    };
    inline CTree( const CFunc *f_in, PTree a0, PTree a1, PTree a2 )
        : typ(_OP), narg(3), fun(f_in) {
        arg[0] = a0; arg[1] = a1; arg[2] = a2;
        has_farg = a0->has_farg || a1->has_farg || a2->has_farg;
        if ( fun->narg!=3 )
            throw string( "PROG ERR: CTree(3) with conflicting #arg" );
    };
    inline CTree( PRef& _ref )
        : typ(_REF), fun(0), ref(_ref), has_farg(false) {};
    inline CTree( PRef& _ref, PTree a0 )
        : typ(_FRF), fun(0), ref(_ref) {
        arg[0] = a0;
        has_farg = a0->has_farg;
    };
    inline CTree( const char which ){
        if        ( which=='t' ) { // function argument
            typ = _FARG;
            has_farg = true;
        } else if ( which=='d' ) { // Dirac delta function centered at 0
            typ = _DIRAC;
            has_farg = true; // function argument is implicit !
            arg[0] = PTree( new CTree( 0.0 ) );
            narg = 1;
        } else
            throw string( "invalid tree type in char-constructor" );
    };
    inline CTree( const char which, const PTree& a0 ){
        if ( which=='d' ) { // Dirac delta function centered at a0
            typ = _DIRAC;
            if( a0->has_farg )
                throw string( "shift of Dirac delta must not contain t" );
            arg[0] = a0;
            narg = 1;
            has_farg = true; // function argument is implicit !
        } else
            throw string( "invalid tree type in char-constructor" );
    };

    inline CTree( const char which, const PTree& a0, const PTree& a1 ){
        if ( which=='c' ) { // convolution
            typ = _CONV;
            has_farg = a0->has_farg;
            arg[0] = a0;
            arg[1] = a1;
            narg = 2;
        } else
            throw string( "invalid tree type in char-constructor" );
    };

    void tree_point_val( double *ret, uint k=(uint)-1,
                         uint j=(uint)-1, uint i=(uint)-1 ) const;
    void tree_vec_val( vector<double> *ret,
                       uint k=(uint)-1, uint j=(uint)-1 ) const;
    void tree_uival( uint *ret, const CContext *ctx ) const;
    void tree_curve_val_vec( vector<double> *ret, const vector<double>& farg,
                             uint k, uint j ) const;
    double tree_curve_val_sca( const double farg, uint k, uint j ) const;
    void coord( CCoord *ret, uint k=(uint)-1 ) const;
    uint npar() const;
    string tree_info() const;
    inline const char* tree_cstr() const {
        return tree_info().c_str(); };

    void tree_val( CTOut *ret, const CContext *ctx ) const;
    static uint setConvK( uint k );
    static uint setConvJ( uint k, uint j, uint kconv );
 private:
    void npar_exec( uint *np ) const;
};


//! Direct access from main loop.

namespace NCalc {
    void Calculator();
};
