//**************************************************************************//
//* FRIDA: fast reliable interactive data analysis                         *//
//* olm.cpp: online memory                                                 *//
//* (C) Joachim Wuttke 1990-, v2(C++) 2001-                                *//
//* http://frida.sourceforge.net                                           *//
//**************************************************************************//

#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <algorithm>
using namespace std;

#include "mystd.h"
#include "olm.h"
#include "asi.h"

int err;


//***************************************************************************//
//* class COlo                                                              *//
//* generic online objects (data or curve files)                            *//
//***************************************************************************//


void COlo::SaveHeader( FILE *fd, bool isdata )
// written for file format y08
{
    string out;
    fprintf( fd, "Meta:\n"
             "  format:  frida/y08 for yaml1\n"
             "  type:    %s\n",
             isdata ? "generic tabular data" : "frida2 curve" );

    fprintf( fd, "History:\n" );
    for ( uint i=0; i<lDoc.size(); i++ )
        fprintf( fd, "- %s\n", yaml(lDoc[i]).c_str() );
    out = string( "saved as " ) + name + " on " +
        mystd::myasctime(mystd::justnow(),0);
    fprintf( fd, "- \%s\n", yaml( out ).c_str() );

    fprintf( fd, "Coord:\n" );
    fprintf( fd, "  x:\n    name: %s\n    unit: %s\n", 
             yaml(xco.name).c_str(), yaml(xco.unit).c_str() );
    fprintf( fd, "  y:\n    name: %s\n    unit: %s\n", 
             yaml(yco.name).c_str(), yaml(yco.unit).c_str() );
    for ( uint i=0; i<ZCo.size(); i++ )
        fprintf( fd, "  z%u:\n    name: %s\n    unit: %s\n", 
                 i, yaml(ZCo[i].name).c_str(), yaml(ZCo[i].unit).c_str() );

    fprintf( fd, "Param:\n" );
    for ( uint i=0; i<RPar.size(); i++ )
        fprintf( fd, "  -\n    name:  %s\n    unit:  %s\n    value: %20.15g\n",
                 yaml(RPar[i].Co.name).c_str(),
                 yaml(RPar[i].Co.unit).c_str(),
                 RPar[i].val );
}


//! Write header in format ASCII-01

void COlo::SaveHeader_a01(FILE *fd)
{
    fprintf( fd, "fil %s\n", name.c_str() );
    lDoc.push_back("saved as "+name+" on "+
                   mystd::myasctime(mystd::justnow(),0));
    for ( uint i=0; i<lDoc.size(); i++ )
        fprintf( fd, "doc %s\n", lDoc[i].c_str() );

    fprintf( fd, "x   %s\n", xco.fstr().c_str() );
    fprintf( fd, "y   %s\n", yco.fstr().c_str() );
    for ( uint i=0; i<ZCo.size(); i++ )
        fprintf( fd, "z%-2u %s\n", i, ZCo[i].fstr().c_str() );
    for ( uint i=0; i<RPar.size(); i++ )
        fprintf( fd, "rpa %20.15g %s\n",
                 RPar[i].val, RPar[i].Co.fstr().c_str() );
}


//***************************************************************************//
//* class COld                                                              *//
//* online data files                                                       *//
//***************************************************************************//


COld::COld( class COlc const* c )
{
    name = c->name;
    lDoc = c->lDoc;
    xco = c->xco;
    yco = c->yco;
    ZCo = c->ZCo;
    RPar = c->RPar;
}

//! Save bulk of online file in one of several formats.

void COld::SaveBody( FILE *fd, string fmt )
{
    if( fmt=="y08" ){
        fprintf( fd, "# %zu tables\n", VS.size() );
        fprintf( fd, "Tables:\n", VS.size() );
        for( size_t j=0; j<VS.size(); j++ ){
            fprintf( fd, "  - # table %zu\n", j );
            VS[j].Save( fd, fmt );
        }
    } else if( fmt=="a01" ){
        fprintf( fd, "#j  %zu\n", VS.size() );
        for( size_t j=0; j<VS.size(); j++ ){
            fprintf( fd, "scn %zu\n", j );
            VS[j].Save( fd, fmt );
        }
    } else if( fmt=="csv" ){
        for( size_t j=0; j<VS.size(); j++ ){
            for( size_t i=0; i<VS[j].size(); ++i ){
                fprintf( fd, "%g\t", VS[j].y[i] );
            }
            fprintf( fd, "\n" );
        } 
    } else if( fmt=="tab" ){
        for( size_t j=0; j<VS.size(); j++ ){
            for( size_t iz=0; iz<ZCo.size(); ++iz ){
                if( iz>0 )  fprintf( fd, "\t" );
                fprintf( fd, "%g", VS[j].z[iz] );
            }
            fprintf( fd, "\n" );
            for( size_t i=0; i<VS[j].size(); ++i ){
                fprintf( fd, "%g\t%g\n", VS[j].x[i], VS[j].y[i] );
            }
            fprintf( fd, "\n" );
        } 
    } else if( fmt=="lis" ){ // sozusagen außer Betrieb
        for( size_t j=0; j<VS.size(); j++ ){
            VS[j].Save( fd, fmt );
            if( j<VS.size()-1 )
                fprintf(fd, "\n", j); // ??? // empty line separates scans
        } 
    } else
        throw string( "SaveBody: invalid data format" );
}


uint COld::nPts( uint j ) const {
    if (!nScan() || j>=nScan())
        throw string( "PROGRAM ERROR: nPts called with invalid j="+strg(j) );
    return VS[j].size();
}


//! Return number of points if it is the same for all scans. Else return 0.

uint COld::nPts() const {
    if (!nScan())
        throw string( "PROGRAM ERROR: nPts(void) called while nScan=0" );
    uint np = VS[0].size();
    for (uint j=0; j<nScan(); ++j)
        if (VS[j].size()!=np) return 0;
    return np;
}


//***************************************************************************//
//* class COlc                                                              *//
//* online curve files                                                      *//
//***************************************************************************//


int COlc::SaveBody(FILE *fd)
// written for file format ASCII-01
{
    fprintf(fd, "#j  %zu\n", VC.size());
    for (size_t j=0; j<VC.size(); j++) {
        fprintf(fd, "scn %zu\n", j);
        if ((err=VC[j].Save(fd))) return err;;
    }
    return 0;
}

//***************************************************************************//
//* class CEle                                                              *//
//* pointer to an online file                                               *//
//***************************************************************************//

void CEle::Save( FILE *fd, string fmt ) 
{
    if( fmt=="y08"){
        ((COlo*) Ptr)->SaveHeader( fd, Typ==DATA );

        if        (Typ==DATA) {
            ((COld*) Ptr)->SaveBody( fd, fmt );
        } else if (Typ==CURVE) {
            ((COlc*) Ptr)->SaveBody( fd );
        } else 
            throw string( "FATAL: invalid Typ in Save" );

        ((COlo*) Ptr)->as_on_disk = true;

    } else if( fmt=="a01"){
        fprintf(fd, "ASCII-01 JWu\n");

        if     (Typ==CURVE)
            fprintf(fd, "typ C\n");
        else if(Typ==DATA)
            fprintf(fd, "typ D\n");
        else
            throw string( "FATAL: invalid Typ in Save" );

        ((COlo*) Ptr)->SaveHeader_a01(fd);

        if        (Typ==DATA) {
            ((COld*) Ptr)->SaveBody( fd, fmt );
        } else if (Typ==CURVE) {
            ((COlc*) Ptr)->SaveBody( fd );
        } else 
            throw string( "FATAL: invalid Typ in Save" );
        
        fprintf(fd, "eof\n");
        ((COlo*) Ptr)->as_on_disk = true;

    } else if( ( fmt=="csv" || fmt=="tab" ) && Typ==DATA ) {
        ((COld*) Ptr)->SaveBody( fd, fmt );

    } else
        throw string( "Unknown format" );
};


//! OBSOLETE ???  ( obsoleted by File_io ? )

void CEle::Load( FILE *fd, string fshort )
{
    throw string( " sorry, in oct07 JWu thought Ele::Load was obsolete" );

    COlc C;
    COld D;
    CScan DS;
    CCurve CS;
    CCoord co;
    uint iz, nj=0, j;
    string lin, test;

    if (mystd::freadln(fd, &lin)<=0)
        throw string( "! file empty or starting with empty line" );
    if (lin!=string("ASCII-01 JWu"))
        throw "! unknown file format: " + lin;
    if (mystd::freadln(fd, &lin)<=0)
        throw string( "! unexpected empty line\n");
    if (lin.substr(0,4)!=string("typ "))
        throw string( "! missing \"typ\" entry\n");
    if        (lin.substr(4)=="D") {
        Typ = DATA;
        Ptr = &D;
    } else if (lin.substr(4)=="C") {
        Typ = CURVE;
        Ptr = &C;
    } else
        throw "FATAL: invalid type: " + lin.substr(4);

    P()->name = fshort;

    while (1) {
        if (mystd::freadln(fd, &lin)<0)
            throw string( "! unexpected end-of-file" );
        if        (lin=="eof") {
            NOlm::OloAdd(P())->P()->as_on_disk = true;
            return; // regular exit
        } else if (lin.substr(0,4)=="fil ") {
            if (lin.substr(4)!=fshort) {
                printf("! warning: file %s renamed from %s\n",
                       fshort.c_str(), lin.substr(4).c_str());
            }
        } else if (lin.substr(0,4)=="doc ") {
            P()->lDoc.push_back(lin.substr(4));
        } else if (lin.substr(0,4)=="x   ") {
            P()->xco.load(lin.substr(4));
        } else if (lin.substr(0,4)=="y   ") {
            P()->yco.load(lin.substr(4));
        } else if (lin[0]=='z') {
            if (sscanf(lin.substr(1,3).c_str(), "%u", &iz)!=1)
                throw string( "z coordinates must be numbered" );
            if (iz!=P()->ZCo.size())
                throw string( "z coordinates must be sorted" );
            co.load(lin.substr(4));
            P()->ZCo.push_back(co);
        } else if (lin.substr(0,4)=="#j  ") {
            if (sscanf(lin.substr(4).c_str(), "%u", &nj)!=1)
                throw string( "after label #j no no. of scans" );
        } else if (lin.substr(0,4)=="scn ") {
            if (sscanf(lin.substr(4).c_str(), "%u", &j)!=1) {
                throw string( "! number of scan not given\n");
            } else if (j!=nScan()) {
                printf("scan %u, label %u\n", nScan(), j);
                throw string( "inconsistency in input file" );
            }
            if        (Typ == DATA) {
                DS.Load( fd, P()->ZCo.size() );
                D.VS.push_back(DS);
            } else if (Typ == CURVE) {
                CS.Load( fd, P()->ZCo.size() );
                C.VC.push_back(CS);
            } else 
                throw string( "FATAL: invalid Typ" );
        } else
            throw "! Unexpected label " + lin.substr(0,4);
    }
}


//***************************************************************************//
//* namespace NOlm                                                          *//
//***************************************************************************//

namespace NOlm { // on-line memory (in olm.h predefined as extern)
    // the data:
    vector<class CEle> MOM;

    // the current selection:
    class CList FSel;
    class CList FSelOld;
    bool overwrite;
};


//***************************************************************************//
//* NOlm: basic administration of online memory                             *//
//***************************************************************************//

//! Clear file selection.
//  To be called after 'eiter' and before creation of new files.

void NOlm::SelNew(void)
{
    FSelOld = FSel;
    FSel = CList();
}

//! Go back to previous file selection.

void NOlm::SelPop(void)
{
    FSel = FSelOld;
    FSelOld = CList();
}

//! Append copy of file *olo to online memory.

CEle* NOlm::OloAdd( const COlo *olo )
{
    CEle e;
    if        ( const COld* olf = dynamic_cast<const COld*>(olo) ) {
        COld *p = new COld;
        *p = *olf;
        e = CEle( p );
    } else if ( const COlc* olc = dynamic_cast<const COlc*>(olo) ) {
        COlc *p = new COlc;
        *p = *olc;
        e = CEle( p );
    } else
        throw string("PROGAMM ERROR: invalid on-line object in OloDup");
    e.P()->as_on_disk = false;
    MOM.push_back( e );
    FSel.append( MOM.size()-1 );
    return &( MOM.back() );
}

//! Remove selected files from online memory.

void NOlm::OlfDel()
{
    for (uint i = FSel.size()-1; i!=(uint)-1; --i) {
        MOM[FSel.V[i]].predelete();
        MOM.erase(MOM.begin()+FSel.V[i]);
    }
    SelPop();
    SelPop();
    SelNew();
    if( MOM.size() > 0 )
        FSel = CList(0,MOM.size()-1);
    else
        FSel = CList();
}

//! Append copy of selected files to online memory.

void NOlm::OlfCopy()
{
    CEle *e;
    IterateEle eiter;
    SelNew();
    while ((e=eiter()))
        OloAdd(e->P());
}

//! Move files to end of online memory.

void NOlm::OlfExchange()
{
    CEle *e;
    IterateEle eiter;
    FSelOld = FSel;
    FSel = CList();
    while ((e=eiter()))
        OloAdd(e->P());
    for (uint i = FSelOld.size()-1; i!=(uint)-1; --i) {
        MOM[FSelOld.V[i]].predelete();
        MOM.erase(MOM.begin()+FSelOld.V[i]);
    }
    for ( uint i=0; i<FSelOld.size(); ++i )
        FSel.V[i] -= FSelOld.size();
}

//! For each file in selection, set fno, fin, fout, and call exec(...).

void NOlm::OlfLoop(void(*exec)(int fno, COld *fin, COld *fout))
{
    IterateD fiter;
    if (!overwrite) SelNew();
    COld *fin, ftmp, *fout;
    int fno;
    while((fin=fiter())) {
        fno = fiter.SelNo();
        cout << "-> operate on file " << fno << "\n";
        OUT_TO_IN( overwrite, fin, ftmp, fout );

        exec(fno, fin, fout);

        if(!(overwrite)) OloAdd(&ftmp);
    }
}


//***************************************************************************//
//* NOlm: scan selection                                                    *//
//***************************************************************************//

void NOlm::JSelAsk( string quest, string *jSel, CList *jLis )
{
    if ( FSel.size()<=0 )
        throw string( "query scan selection: no files to operate on" );

    uint nJ = MOM[FSel.V[0]].nScan();

    IterateEle iter;
    CEle *e;
    while ( (e=iter()) )
        if ( e->nScan()!=nJ )
            throw string( "files have different numbers of scans" );

    if ( nJ==1 )
        *jLis = CList("0");
    else {
        *jSel = wask( quest, *jSel );
        *jLis = CList ( *jSel, 0, nJ-1 );
    }
}
